/*
 * Copyright (C) 2011 Canonical Ltd
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 3 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Authored by Alex Launi <alex.launi@canonical.com>
 *
 */

using GLib;

namespace Unity.MusicLens {
 	
  public class BansheeScopeProxy : Object
  {
    public Unity.Scope Scope { get; private set; }
    
    private bool is_dirty;
    private LensSearch? previous_search;
    private BansheeCollection collection;
    
    construct {
      Scope = new Unity.Scope ("/com/canonical/unity/scope/banshee");
      Scope.search_in_global = true;
      Scope.activate_uri.connect (activate);
      
      try {
	collection = new BansheeCollection ();
      } catch (DatabaseError error) {
	printerr ("Failed to open the Banshee collection database\n");
	return;
      }
      
      is_dirty = true;
      previous_search = null;
      
      /* Listen for filter changes */
      Scope.filters_changed.connect (
				     () => {          
				       if (Scope.active_search != null)
					 {
					   is_dirty = true;
					   Scope.notify_property ("active-search");
					 }
				     }
				     );
      
      /* Listen for changes to the lens entry search */
      Scope.notify["active-search"].connect (
					     (obj, pspec) => {
					       var search = Scope.active_search;
					       
					       if (!(is_dirty || search_has_really_changed (previous_search, search)))
						 return;
					       
					       is_dirty = false;
					       
					       update_search_async.begin (search);
					       previous_search = search;
					     }
					     );
      
      /* Listen for changes to the global search */
      Scope.notify["active-global-search"].connect (
						    (obj, pspec) => {
						      
						      var search = Scope.active_global_search;
						      
						      if (search_is_invalid (search))
							return;
						      
						      if (!search_has_really_changed (previous_search, search))
							return;
						      
						      update_global_search_async.begin (search);
						      previous_search = search;
						    }
						    );
    }
    
    /**
     * Tells banshee to play the selected uri(s)
     */
    public Unity.ActivationResponse activate (string uri)
    {
      string[] exec = {"banshee", "--play"};
      
      try {
        if (Uri.parse_scheme (uri) == "album")
          {
            debug (@"searching for tracks for $uri");
            string[] split = uri.split ("/");
            string artist = split[2];
            string title = split[3];
	    
            Album album = new Album ();
            album.Artist = artist;
            album.Title = title;
            // FIXME there must be a better way..
            foreach (string track in collection.get_track_uris (album))
              exec += track;
          }
        else
          {
            exec += uri;
          }
	
        exec += null;
	
        debug (@"Spawning banshee --play $uri");
        Process.spawn_async (null, 
			     exec,
			     null, 
			     SpawnFlags.SEARCH_PATH,
			     null, 
			     null);
        return new Unity.ActivationResponse (Unity.HandledType.HIDE_DASH);
      } catch (SpawnError error) {
        warning ("Failed to launch URI %s", uri);
        return new Unity.ActivationResponse (Unity.HandledType.NOT_HANDLED);
      }
    }
    
    /* based on the filters(?) get the default results to show when the music lens is
     * just opened without filters, like when someone hits Super+M
		 */
    private async void update_without_search_async (LensSearch search)
    {
      perform_search_async.begin (search, 100);
    }

    /**
     * results for a global search from just hitting Super. here we want to return
     * a smaller number of results with 0 filters.
     */
    private async void update_global_search_async (LensSearch search)
    {
      var results_model = Scope.global_results_model;
      
      if (search_is_invalid (search))
	{
	  results_model.clear ();
	  return;
	}
      
      perform_search_async.begin (search, 20);
    }
    
    private async void update_search_async (LensSearch search)
    {
      // just pretend like there's no search
      if (search_is_invalid (search))
	{
	  update_without_search_async.begin (search);
	  return;
	}
      
      perform_search_async.begin (search, 50);
    }
    
    private async void perform_search_async (LensSearch search, int max_results)
    {
      var results_model = Scope.results_model;
      
      // here we get the filters. I'm thinking the best way would be to make a GenreFilter
      // and a DateRange filter class that would convert what's in the dash filters to SQL
      // and then pass a LinkedList<BansheeFilter> into collecton.search and it will walk
      // through the filters in it and take the sql from them and pass it to the query
      List<FilterParser> filters = new List<FilterParser> ();
      Filter filter = Scope.get_filter ("genre");
      if (filter.filtering)
	filters.append (new FilterParser (filter));
      
      filter = Scope.get_filter ("decade");
      if (filter.filtering)
	filters.append (new FilterParser (filter));
      
      /* Prevent concurrent searches and concurrent updates of our models,
       * by preventing any notify signals from propagating to us.
       * Important: Remeber to thaw the notifys again! */
      Scope.freeze_notify ();
      
      results_model.clear ();
      collection.search (search, results_model, filters, max_results);
      
      /* Allow new searches once we enter an idle again.
       * We don't do it directly from here as that could mean we start
       * changing the model even before we had flushed out current changes
       */
      Idle.add (() => {
	  Scope.thaw_notify ();
	  return false;
	});
      
      search.finished ();
    }
    
    private bool search_is_invalid (LensSearch? search)
    {
      /* This boolean expression is unfolded as we seem to get
       * some null dereference if we join them in a big || expression */
      if (search == null)
        return true;
      else if (search.search_string == null)
        return true;
      
      return search.search_string.strip() == "";
    }

    private bool search_has_really_changed (LensSearch? old_search,
					    LensSearch? new_search)
    {
      if (old_search == null && new_search == null)
	return false;
      
      string s1, s2;
      
      if (old_search == null)
	{
	    return true;
	}
      else if (new_search == null)
	{
	  s2 = old_search.search_string;
	  if (s2 == null || s2.strip() == "")
	    return false;
	  else
	    return true;
	}
      else
	{
	  s1 = new_search.search_string;
	  s2 = old_search.search_string;
	  if (s1 == null)
	    {
	      if (s2 == null || s2.strip() == "")
		return false;
	      else
		return true;
	    }
	  else if (s2 == null)
	    {
	      if (s1 == null || s1.strip() == "")
		return false;
	      else
		return true;
	    }
	  else
	    return s1.strip () != s2.strip ();
	}
    }
  }
}
