/*
 * Copyright (C) 2011 Canonical Ltd
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 3 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Authored by Alex Launi <alex.launi@canonical.com>
 *
 */

using GLib;

namespace Unity.MusicLens {
  
  public class FilterParser : Object
  {
    private Genre map;
    private Filter _filter;
    
    public string Id { get { return _filter.id; } }

    public FilterParser (Filter filter)
    {
      _filter = filter;
      map = new Genre ();
    }

    public string? parse ()
    {
      if (Id == "genre")
        return parse_genre_filter ();
      else if (Id == "decade")
        return parse_decade_filter ();
      
      return null;
    }

    /* returns a string like CoreTracks.Genre IS 'Rock' OR CoreTracks.Genre IS 'Metal'... */
    public string? parse_genre_filter ()
    {
      StringBuilder builder = new StringBuilder ();
      List<FilterOption> genres = get_all_selected_genre_options ((_filter as CheckOptionFilter).options);

      /* CoreTracks.Genre IS ___ OR CoreTracks.Genre IS ____ ... */
      foreach (FilterOption genre in genres)
      {
        string id = genre.id;
        if (genre.id == null)
          continue;

        foreach (string alt in map.get_genre_synonyms (id))
        {
          builder.append ("(CoreTracks.Genre LIKE '%");
          builder.append (alt);
          builder.append ("""%' ESCAPE '\' AND CoreTracks.Genre IS NOT NULL) OR """);
        }
      }
      builder.truncate (builder.len - (" OR ".length));
      builder.prepend ("(");
      builder.append (")");
     
      return builder.str;
    }

    /* returns a string like CoreTracks.Year >= LOWER_BOUND AND CoreTracks.Year < UPPER_BOUND */
    public string? parse_decade_filter ()
    {
      int start_year, end_year;

      MultiRangeFilter range_filter = _filter as MultiRangeFilter;
      FilterOption start = range_filter.get_first_active ();
      FilterOption end = range_filter.get_last_active ();

      start_year = int.parse (start.id);
      end_year = int.parse (end.id);
      
      // bump the end_year up to the beginning of the next range
      int end_index = range_filter.options.index (end);
      // if we have already selected the last option
      if (end_index == range_filter.options.length () - 1)
        end_year += 10;
      else
        end_year = int.parse (range_filter.options.nth_data (end_index + 1).id);

      return "(CoreTracks.Year >= %d AND CoreTracks.Year < %d) ".printf (start_year, end_year);
    }

    private List<FilterOption> get_all_selected_genre_options (List<FilterOption> options)
    {
      var active = new List<FilterOption> ();
      foreach (FilterOption option in options)
      {
        if (option.active)
          active.append (option);
      }

      return active;
    }
  }
}
